"""
amof.coordination
---
Submodule containing coordination analysis related methods
Import entire core content
Called by every submodule in coordination
"""

# from . import core # to import core when importing submodule
from .core import *